package fsl.depict.netcdf;

import java.io.File;
import java.util.ArrayList;
import fsl.util.ConfigTable;

/**
 * The formatting rule set, encapsulating a set of rules that
 * dictate how the data within a Netcdf file is to be formatted
 * when converted into a product.
 *
 * @author Christopher Golden
 */
public class FormattingRuleSet {


	// Private Static Constants

	/**
	 * String parameter type.
	 */
	private static final int STRING_TYPE = 0;

	/**
	 * Integer parameter type.
	 */
	private static final int INTEGER_TYPE = 1;

	/**
	 * Floating point parameter type.
	 */
	private static final int FLOATING_POINT_TYPE = 2;

	/**
	 * Array of strings indicating placement of a parameter;
	 * the index of each corresponds to the appropriate value
	 * for the <code>placement</code> attribute of <code>
	 * FormattingRule</code>.
	 *
	 * @see fsl.depict.netcdf.FormattingRule
	 */
	private static final String[] PLACEMENT_STRINGS = { "N", "NE", "E", "SE", "S",
														"SW", "W", "NW", "C" };


	// Private Variables

	/**
	 * List of rules of type <code>FormattingRule</code>.
	 */
	private ArrayList<FormattingRule> rules = new ArrayList<FormattingRule>();

	/**
	 * Boolean indicating whether or not there is a center
	 * data point.
	 */
	private boolean centerData = false;


	// Public Constructors

	/**
	 * Construct a standard instance.
	 *
	 * @param     file File upon which to base the rule set.
	 * @exception Exception if the file incorrectly specifies
	 *            the rules.
	 */
	public FormattingRuleSet(File file) throws Exception {

		// Read through the rule set file as a table of N
		// lines, one per rule.  Each rule consists of a
		// number of different elements.  If an error occurs,
		// throw an exception.
		boolean[] placementUsed = new boolean[PLACEMENT_STRINGS.length];
		ConfigTable configTable = new ConfigTable(file);
		for (int j = 0; j < configTable.getLineCount(); j++) {

			// Make sure that the correct number of elements
			// have been provided.
			String[] line = configTable.getLine(j);
			if (line == null)
				continue;
			else if (line.length != 6)
				throw new Exception("line " + (j + 1) + ": wrong number args: " +
									"should be \"netcdfID placement format " +
									"conversion minValue maxValue\"");

			// Get the identifier of the parameter.
			String identifier = line[0];

			// Get the placement of the parameter.
			int placement;
			for (placement = 0; placement < PLACEMENT_STRINGS.length; placement++)
				if (PLACEMENT_STRINGS[placement].equals(line[1].toUpperCase()))
					break;
			if (placement == PLACEMENT_STRINGS.length)
				throw new Exception("line " + (j + 1) + ": illegal placement " +
									"string \"" + line[1] + "\"");
			else if (placementUsed[placement])
				throw new Exception("line " + (j + 1) + ": duplicate use of " +
									"placement \"" + line[1] + "\"");
			placementUsed[placement] = true;

			// Get the format and type of the parameter.
			line[2] = line[2].toUpperCase();
			int type, digitsAfter = 0;
			if (line[2].equals("S"))
				type = STRING_TYPE;
			else if (line[2].equals("I"))
				type = INTEGER_TYPE;
			else if (line[2].endsWith("F")) {
				type = FLOATING_POINT_TYPE;
				line[2] = line[2].substring(0, line[2].length() - 1);
				try {
					digitsAfter = Integer.parseInt(line[2]);
					if (digitsAfter < 1)
						throw new Exception();
				} catch (Exception e) {
					throw new Exception("bad digits-following-decimal-point " +
										"specifier \"" + line[2] + "\"");
				}
			} else
				throw new Exception("line " + (j + 1) + ": illegal format type \"" +
									line[2] + "\"");

			// Get the conversion of the parameter.
			// IGNORE SPECIFIED STRING FOR NOW.
			int conversion = 0;

			// Get the minimum and maximum values for
			// the parameter if it is numerical.
			int intMin = 0, intMax = 0;
			double floatMin = 0.0, floatMax = 0.0;
			try {
				if (type == INTEGER_TYPE) {
					intMin = Integer.parseInt(line[4]);
					intMax = Integer.parseInt(line[5]);
					if (intMin > intMax)
						throw new Exception();
				} else if (type == FLOATING_POINT_TYPE) {
					floatMin = Double.parseDouble(line[4]);
					floatMax = Double.parseDouble(line[5]);
					if (floatMin > floatMax)
						throw new Exception();
				}
			} catch (Exception e) {
				throw new Exception("bad minimum-maximum value specifier \"" +
									line[4] + " " + line[5] + "\"");
			}

			// Construct a rule as specified.
			FormattingRule rule;
			if (type == STRING_TYPE)
				rule = new FormattingRule(identifier, placement);
			else if (type == INTEGER_TYPE)
				rule = new IntegerFormattingRule(identifier, placement, conversion,
												 intMin, intMax);
			else
				rule = new FloatingPointFormattingRule(identifier, placement,
													   conversion, digitsAfter,
													   floatMin, floatMax);
			rules.add(rule);
		}

		// Remember whether or not there is a center
		// data point.
		centerData = placementUsed[FormattingRule.CENTER];
	}


	// Public Methods

	/**
	 * Get the number of rules in this rule set.
	 *
	 * @return Number of rules in this rule set.
	 */
	public int getSize() {
		return rules.size();
	}

	/**
	 * Get the rule at the specified index.
	 *
	 * @param  index Index of the rule to be fetched.
	 * @return Rule that was requested.
	 */
	public FormattingRule getRule(int index) {
		return rules.get(index);
	}

	/**
	 * Find out whether or not there is a center
	 * data point to be plotted.
	 *
	 * @return True if there is a center data point
	 *         to be plotted, false otherwise.
	 */
	public boolean centerDataPointExists() {
		return centerData;
	}
}
