package fsl.depict.netcdf;

import java.io.File;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import ucar.ma2.Array;
import ucar.ma2.Index;
import ucar.nc2.Attribute;
import ucar.nc2.Dimension;
import ucar.nc2.NetcdfFile;
import ucar.nc2.Variable;
import fsl.util.DGMArray;

/**
 * Application to be used to convert Netcdf into products that
 * may be displayed by FXC.  The application takes three command
 * line arguments: the first is the path and name of the Netcdf
 * file to be converted, the second the path and name of the file
 * to be created to store the resulting product, and the third
 * the path and name of the file containing the formatting rules
 * that will be used to dictate which parameters the product
 * should display and how they should be displayed.
 *
 * @author Christopher Golden
 */
public class NetcdfConverter {


	// Private Static Constants

	/**
	 * Name of the variable which itself holds the names of the
	 * latitude and longitude variables within the Netcdf file.
	 */
	private static final String LAT_LONG_VARIABLE_META_NAME = "latLongVars";

	/**
	 * Default latitude variable name.
	 */
	private static final String LATITUDE_DEFAULT_VARIABLE_NAME = "staLat";

	/**
	 * Default longitude variable name.
	 */
	private static final String LONGITUDE_DEFAULT_VARIABLE_NAME = "staLon";

	/**
	 * X coordinates of the glyph to be drawn at a station's
	 * location within the DGM product file.
	 */
	private static final short[] GLYPH_X_COORDS = { -2, 0, 2, 0, -2, 2, 0, 0 };

	/**
	 * Y coordinates of the glyph to be drawn at a station's
	 * location within the DGM product file.
	 */
	private static final short[] GLYPH_Y_COORDS = { 0, 2, 0, -2, 0, 0, 2, -2 };

	/**
	 * X coordinates of the locations at which text should
	 * be drawn for the various placements, indexed by the
	 * values of the directional constants defined in <code>
	 * FormattingRule</code>.
	 *
	 * @see fsl.depict.netcdf.FormattingRule
	 */
	private static final short[] PARAM_X_COORDS = { 0, 7, 12, 7, 0, -7, -12, -7, 0 };

	/**
	 * Y coordinates of the locations at which text should
	 * be drawn for the various placements, indexed by the
	 * values of the directional constants defined in <code>
	 * FormattingRule</code>.
	 *
	 * @see fsl.depict.netcdf.FormattingRule
	 */
	private static final short[] PARAM_Y_COORDS = { 20, 10, 0, -10, -20, -10, 0, 10, 0 };


	// Public Static Methods

	/**
	 * Entry point for the application.
	 *
	 * @param args Array of arguments specified by the user
	 *             on the command line.
	 */
	public static void main(String[] args) {

		// Make sure that there are three arguments provided.
		// The first should be the path (or URL) of the Netcdf
		// file to be converted, the second the name of the
		// resulting file, and the third the name of the file
		// that holds the rules to be used for formatting the
		// resulting file.
		if (args.length != 3) {
			System.err.println("Wrong args: should be \"sourceNetcdfFile " +
							   "destinationFile formatRulesFile\".");
			System.exit(1);
		}
		try {

			// Create a formatting rule set based upon the
			// specified formatting rules file.
			FormattingRuleSet rules = new FormattingRuleSet(new File(args[2]));

			// Open the Netcdf file for reading.
			NetcdfFile file = null;
			if (args[0].startsWith("http://"))
				file = new NetcdfFile(new URL(args[0]));
			else
				file = new NetcdfFile(args[0]);

			// Determine the names of the latitude and longitude
			// variables in the file, using defaults if no global
			// attribute can be found to help.
			String latVarName, longVarName;
			Attribute metaName = file.findGlobalAttribute(LAT_LONG_VARIABLE_META_NAME);
			if (metaName != null) {
				String string = (String) metaName.getValue();
				int commaIndex = string.indexOf(',');
				latVarName = string.substring(0, commaIndex - 1);
				longVarName = string.substring(commaIndex + 1);
			} else {
				latVarName = LATITUDE_DEFAULT_VARIABLE_NAME;
				longVarName = LONGITUDE_DEFAULT_VARIABLE_NAME;
			}

			// Find the latitude and longitude variables, throw-
			// ing an error if one or both are not found.
 			Variable latVar = file.findVariable(latVarName);
			if (latVar == null)
				throw new Exception("No latitude variable found.");
			Variable longVar = file.findVariable(longVarName);
			if (longVar == null)
				throw new Exception("No longitude variable found.");

			// Create a DGM file that displays each station's
			// name at the proper lat-long coordinates.
			DGMArray dgm = new DGMArray();
			dgm.setFont(DGMArray.FONT_7_BY_9_ASCII);

			// Get the data for the variables, and if they do
			// not all have the same primary dimension, throw
			// an error.
			Array latData = latVar.read();
			int[] latShape = latData.getShape();
			Index latIndex = latData.getIndex();
			Array longData = longVar.read();
			int[] longShape = longData.getShape();
			Index longIndex = longData.getIndex();
			if (latShape[0] != longShape[0])
				throw new Exception("Dimensions of lat-long variables do not match up.");
			Variable[] variables = new Variable[rules.getSize()];
			Array[] data = new Array[rules.getSize()];
			int[][] shapes = new int[rules.getSize()][];
			Index[] indices = new Index[rules.getSize()];
			for (int j = 0; j < rules.getSize(); j++) {
				FormattingRule rule = rules.getRule(j);
				variables[j] = file.findVariable(rule.getIdentifier());
				if (variables[j] == null)
					throw new Exception("No variable \"" + rule.getIdentifier() +
										"\" found.");
				data[j] = variables[j].read();
				shapes[j] = data[j].getShape();
				indices[j] = data[j].getIndex();
				if (shapes[j][0] != latShape[0])
					throw new Exception("Dimensions of variable \"" +
										rule.getIdentifier() + "\" does not " +
										"match that of lat-long variables.");
			}

			// Iterate through the stations, putting a glyph
			// at each one's latitude and longitude, surrounded
			// by the values of the variables specified in the
			// rule set.
			for (int j = 0; j < latShape[0]; j++) {

				// Get the station latitude and longitude, and
				// set relative addressing in the DGM array with
				// this lat-long as the center point.
				latIndex.set(j);
				float latitude = latData.getFloat(latIndex);
				longIndex.set(j);
				float longitude = longData.getFloat(longIndex);
				dgm.setRelativeAddressing(DGMArray.RELATIVE_ADDRESSING_NORMAL,
										  (short) ((longitude * (float) 60.0) +
												   (float) 0.5),
										  (short) ((latitude * (float) -60.0) +
												   (float) 0.5));

				// Add the station glyph to the array if there
				// is no center point parameter.
				if (rules.centerDataPointExists() == false)
					dgm.addLinkedVectors((short) GLYPH_X_COORDS.length,
										 GLYPH_X_COORDS, GLYPH_Y_COORDS);

				// Go through the list of rules dictating what
				// parameters should be displayed, and for each
				// one, add the necessary drawing commands to
				// the DGM array to display the parameter.
				for (int k = 0; k < rules.getSize(); k++) {
					String paramText = null;
					FormattingRule rule = rules.getRule(k);
					indices[k].set(j);

					// Handle the creation of the text string
					// differently depending upon whether the
					// value is formatted as a floating point
					// number, an integer, or a string.
					if (rule instanceof FloatingPointFormattingRule) {

						// Get the value and make sure it is
						// within the min-max boundaries that
						// the rule specifies.
						FloatingPointFormattingRule floatRule =
							(FloatingPointFormattingRule) rule;
						double value = data[k].getDouble(indices[k]);
						if (value < floatRule.getMinimumValue())
							value = floatRule.getMinimumValue();
						else if (value > floatRule.getMaximumValue())
							value = floatRule.getMaximumValue();
						paramText = Double.toString(value);

						// Make sure that the number of digits
						// following the decimal point are
						// less than or equal to the maximum
						// number specified by the rule.
						int pointIndex = paramText.indexOf('.');
						if (pointIndex >= 0) {
							int digitsAfter = floatRule.getDigitsAfter();

							// Round the number (as a string)
							// if the number of digits that
							// follow the decimal point exceeds
							// the allowable number.
							if (((paramText.length() - 1) - pointIndex) > digitsAfter) {
								char cutDigit = paramText.charAt(pointIndex +
																 digitsAfter + 1);

								// If the number of digits
								// allowed after the decimal
								// point is 0, do the rounding
								// one way; otherwise, do it
								// another.
								if (digitsAfter == 0) {
									if (cutDigit > '4')
										paramText =
											paramText.substring(0, pointIndex - 1) +
											(char) (paramText.
													charAt(pointIndex - 1) + 1);
									else
										paramText = paramText.substring(0, pointIndex);
								} else {
									if (cutDigit > '4')
										paramText =
											paramText.substring(0, pointIndex +
																digitsAfter) +
											(char) (paramText.
													charAt(pointIndex +
														   digitsAfter) + 1);
									else
										paramText = paramText.substring(0, pointIndex +
																		digitsAfter + 1);
								}
							}
						}
					} else if (rule instanceof IntegerFormattingRule) {

						// Get the value and make sure it is
						// within the min-max boundaries that
						// the rule specifies.
						IntegerFormattingRule intRule = (IntegerFormattingRule) rule;
						int value = data[k].getInt(indices[k]);
						if (value < intRule.getMinimumValue())
							value = intRule.getMinimumValue();
						else if (value > intRule.getMaximumValue())
							value = intRule.getMaximumValue();
						paramText = Integer.toString(value);
					} else {

						// Get the characters and assemble
						// them into a string.
						char[] characters = new char[shapes[k][1]];
						int q;
						for (q = 0; q < shapes[k][1]; q++) {
							indices[k].set(j, q);
							characters[q] = data[k].getChar(indices[k]);
							if (characters[q] == (char) 0)
								break;
						}
						paramText = new String(characters, 0, q);
					}

					// Set the text alignment appropriately
					// for the placement of this parameter.
					int placement = rule.getPlacement();
					dgm.setTextAlignment(((placement == FormattingRule.NORTH) ||
										  (placement == FormattingRule.SOUTH) ||
										  (placement == FormattingRule.CENTER) ?
										  DGMArray.TEXT_ALIGNMENT_HORIZONTAL_MIDDLE :
										  (placement < FormattingRule.SOUTH ?
										   DGMArray.TEXT_ALIGNMENT_HORIZONTAL_LEFT :
										   DGMArray.TEXT_ALIGNMENT_HORIZONTAL_RIGHT)),
										 ((placement == FormattingRule.EAST) ||
										  (placement == FormattingRule.WEST) ||
										  (placement == FormattingRule.CENTER) ?
										  DGMArray.TEXT_ALIGNMENT_VERTICAL_MIDDLE :
										  ((placement < FormattingRule.WEST) &&
										   (placement > FormattingRule.EAST) ?
										   DGMArray.TEXT_ALIGNMENT_VERTICAL_BOTTOM :
										   DGMArray.TEXT_ALIGNMENT_VERTICAL_TOP)));
					dgm.addText(PARAM_X_COORDS[placement],
								PARAM_Y_COORDS[placement], paramText);
				}
			}

			// Store the DGM array to the file specified on
			// the command line.
			File destinationFile = new File(args[1]);
			dgm.store(destinationFile.getParent(), destinationFile.getName());
		} catch (Exception e) {
			System.err.println("Exception: " + e);
			e.printStackTrace(System.err);
			System.exit(1);
		}
		System.exit(0);
	}
}
