package fsl.depict.shapefile;

import java.awt.Color;
import java.io.File;
import java.io.Serializable;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import fsl.depict.DepictableDisplayProperties;
import fsl.util.Logger;

/**
 * Class that encapsulates shapefile display properties.
 *
 * @author Christopher Golden
 * @see fsl.depict.shapefile.ShapefileDepictable
 */
public class ShapefileDisplayProperties extends DepictableDisplayProperties {


	// Protected Static Constants

	/**
	 * Maximum number of values a text property may have
     * in order to be included in the property mappings
     * for a shapefile. If -1, there is no limit.
	 */
	private static final int MAX_TEXT_PROPERTY_VALUES = -1;


    // Protected Classes

    /**
     * Basic shapefile property.
     */
    public static abstract class Property implements Serializable, Cloneable {


        // Protected Variables

        /**
         * Name of the property.
         */
        protected String name;


        // Public Constructors

        /**
         * Construct a standard instance.
         *
         * @param name Name of the property.
         */
        public Property(String name) {
            this.name = name;
        }


        // Public Methods

        /**
         * Get the property name.
         *
         * @return Property name.
         */
        public String getName() {
            return name;
        }

        /**
         * Return a clone of this object.
         *
         * @return Clone of this object.
         */
        public Object clone() {
            try {
                return super.clone();
            } catch (Exception e) {
                Logger.logBug("SHOULDN'T HAPPEN.", e);
                return null;
            }
        }

        /**
         * Determine whether the specified property
         * has the same name and type as this one.
         *
         * @param  other Other property to which this
         *               one is to be compared.
         * @return True if the properties have the
         *         same name and type, otherwise
         *         false.
         */
        public abstract boolean hasSameNameAndTypeAs(Property other);

        /**
         * Determine whether the specified property
         * is equivalent to this one in terms of
         * its name, type, and potential values.
         *
         * @param  other Other property to which this
         *               one is to be compared.
         * @return True if the properties are
         *         equivalent as defined above, false
         *         otherwise.
         */
        public abstract boolean isBasicallySameAs(Property other);

        /**
         * Get a list of the property values as
         * displayable strings.
         *
         * @return List of property values as strings
         *         suitable for display.
         */
        public abstract ArrayList<String> getDisplayablePropertyValues();

        /**
         * Get a list of the visibility flags that
         * correspond to property values at the same
         * indices.
         *
         * @return List of visibility flags.
         */
        public abstract ArrayList<Boolean> getVisibilities();

        /**
         * Get a list of the colors that correspond to
         * property values at the same indices.
         *
         * @return List of colors.
         */
        public abstract ArrayList<Color> getColors();

        /**
         * Determine whether the specified index is
         * visible.
         *
         * @param  index Index for which to check the
         *               visibility.
         * @return Visibility of the index.
         */
        public abstract Boolean isVisible(int index);

        /**
         * Set the flag indicating whether the
         * specified index is visible.
         *
         * @param  index   Index for which to set the
         *                 visibility.
         * @param  visible New visibility.
         */
        public abstract void setVisibility(int index, Boolean visible);

        /**
         * Get the color at the specified index.
         *
         * @param  index Index from which to fetch the
         *               color.
         * @return Color.
         */
        public abstract Color getColor(int index);

        /**
         * Set the color at the specified index.
         *
         * @param index Index at which to set the
         *              color.
         * @param color Color to be used.
         */
        public abstract void setColor(int index, Color color);

        /**
         * Incorporate relevant mappings from the
         * specified property into this one. Relevant
         * mappings are those attached to property
         * values that this property also may possess.
         *
         * @param other Other property from which
         *              relevant mappings must be
         *              taken.
         */
        public abstract void incorporateRelevantMappings(Property other);
    }

    /**
     * Shapefile property of the text (CHARACTER) type.
     */
    public static class TextProperty extends Property {


        // Protected Variables

        /**
         * List of property value text strings; these are
         * paired with the colors at the corresponding
         * indices in <code>colors</code>.
         */
        protected String[] values = null;

        /**
         * List of visibility flags; these are paired with
         * the text values at the corresponding indices in
         * <code>values</code>.
         */
        protected Boolean[] visibilityFlags = null;

        /**
         * List of colors; these are paired with the text
         * values at the corresponding indices in <code>
         * values</code>.
         */
        protected Color[] colors = null;


        // Public Constructors

        /**
         * Construct a standard instance with all text
         * values visible and mapped to the same color.
         *
         * @param name   Property name.
         * @param values List of text values.
         * @param color  Starting color for all text
         *               values.
         */
        public TextProperty(String name, String[] values, Color color) {
            super(name);
            this.values = values;
            visibilityFlags = new Boolean[values.length];
            colors = new Color[values.length];
            for (int j = 0; j < values.length; j++) {
                visibilityFlags[j] = Boolean.TRUE;
                colors[j] = color;
            }
        }

        /**
         * Construct a standard instance.
         *
         * @param name     Property name.
         * @param values   List of text values.
         * @param visFlags List of text values.
         * @param colors   List of colors corresponding
         *                 to the text values in
         *                 <code>values</code>.
         */
        public TextProperty(String name, String[] values, Boolean[] visFlags,
                            Color[] colors) {
            super(name);
            this.values = values;
            this.visibilityFlags = visFlags;
            this.colors = colors;
        }


        // Public Methods

        /**
         * Return a clone of this object.
         *
         * @return Clone of this object.
         */
        public Object clone() {
            TextProperty obj = null;
            try {
                obj = (TextProperty) super.clone();
            } catch (Exception e) {
                Logger.logBug("SHOULDN'T HAPPEN.", e);
            }
            obj.values = (String[]) values.clone();
            obj.visibilityFlags = (Boolean[]) visibilityFlags.clone();
            obj.colors = (Color[]) colors.clone();
            return obj;
        }

        /**
         * Get the total number of text value and color
         * pairings.
         *
         * @return Total number of values.
         */
        public int getSize() {
            return values.length;
        }

        /**
         * Get the text value at the specified index.
         *
         * @param  index Index from which to fetch the
         *               text value.
         * @return Text value.
         */
        public String getText(int index) {
            return values[index];
        }

        /**
         * Determine whether the specified index is
         * visible.
         *
         * @param  index Index for which to check the
         *               visibility.
         * @return Visibility of the index.
         */
        public Boolean isVisible(int index) {
            return visibilityFlags[index];
        }

        /**
         * Determine whether the specified text is
         * visible.
         *
         * @param  value Text value for which to get
         *               the visibility.
         * @return Visibility for the text value, or
         *         <code>null</code> if no such value
         *         is found.
         */
        public Boolean getVisibilityForText(String text) {
            for (int j = 0; j < values.length; j++) 
                if (values[j].equals(text))
                    return visibilityFlags[j];
            return null;
        }

        /**
         * Get the color at the specified index.
         *
         * @param  index Index from which to fetch the
         *               color.
         * @return Color.
         */
        public Color getColor(int index) {
            return colors[index];
        }

        /**
         * Get the color for the specified text value.
         *
         * @param  value Text value for which to get
         *               the color.
         * @return Color for the text value, or <code>
         *         null</code> if no such value is found.
         */
        public Color getColorForText(String text) {
            for (int j = 0; j < values.length; j++) 
                if (values[j].equals(text))
                    return colors[j];
            return null;
        }

        /**
         * Determine whether the specified property
         * has the same name and type as this one.
         *
         * @param  other Other property to which this
         *               one is to be compared.
         * @return True if the properties have the
         *         same name and type, otherwise
         *         false.
         */
        public boolean hasSameNameAndTypeAs(Property other) {
            return (name.equals(other.name) && (other instanceof TextProperty));
        }

        /**
         * Determine whether the specified property
         * is equivalent to this one in terms of
         * its name, type, and potential values.
         *
         * @param  other Other property to which this
         *               one is to be compared.
         * @return True if the properties are
         *         equivalent as defined above, false
         *         otherwise.
         */
        public boolean isBasicallySameAs(Property other) {

            // If the name, type, or number of possible
            // values is different, return false.
            if ((name.equals(other.name) == false) ||
                ((other instanceof TextProperty) == false))
                return false;
            TextProperty other2 = (TextProperty) other;
            if (values.length != other2.values.length)
                return false;

            // Iterate through the possible values,
            // ensuring that each is part of the other
            // property's possible-values list as well.
            for (int j = 0; j < values.length; j++) {
                boolean found = false;
                for (int k = 0; k < other2.values.length; k++) {
                    if (values[j].equals(other2.values[k])) {
                        found = true;
                        break;
                    }
                }
                if (found == false)
                    return false;
            }
            return true;
        }

        /**
         * Get a list of the property values as
         * displayable strings.
         *
         * @return List of property values as strings
         *         suitable for display.
         */
        public ArrayList<String> getDisplayablePropertyValues() {
            ArrayList<String> list = new ArrayList<String>();
            for (int j = 0; j < values.length; j++)
                list.add(values[j]);
            return list;
        }

        /**
         * Get a list of the visibility flags that
         * correspond to property values at the same
         * indices.
         *
         * @return List of visibility flags.
         */
        public ArrayList<Boolean> getVisibilities() {
            ArrayList<Boolean> list = new ArrayList<Boolean>();
            for (int j = 0; j < visibilityFlags.length; j++)
                list.add(visibilityFlags[j]);
            return list;
        }

        /**
         * Get a list of the colors that correspond to
         * property values at the same indices.
         *
         * @return List of colors.
         */
        public ArrayList<Color> getColors() {
            ArrayList<Color> list = new ArrayList<Color>();
            for (int j = 0; j < colors.length; j++)
                list.add(colors[j]);
            return list;
        }

        /**
         * Incorporate relevant mappings from the
         * specified property into this one. Relevant
         * mappings are those attached to property
         * values that this property also may possess.
         *
         * @param other Other property from which
         *              relevant mappings must be
         *              taken.
         */
        public void incorporateRelevantMappings(Property other) {

            // Iterate through the other property's
            // mappings, attempting to use each one
            // in turn; any that do not apply because
            // the value does not exist in this
            // property will automatically be ignored.
            TextProperty other2 = (TextProperty) other;
            for (int j = 0; j < other2.values.length; j++) {
                setVisibilityForText(other2.values[j], other2.visibilityFlags[j]);
                setColorForText(other2.values[j], other2.colors[j]);
            }
        }

        /**
         * Set the flag indicating whether the
         * specified index is visible.
         *
         * @param  index   Index for which to set the
         *                 visibility.
         * @param  visible New visibility.
         */
        public void setVisibility(int index, Boolean visible) {
            visibilityFlags[index] = visible;
        }

        /**
         * Set the color at the specified index.
         *
         * @param index Index at which to set the color.
         * @param color New color value to be used.
         */
        public void setColor(int index, Color color) {
            colors[index] = color;
        }

        /**
         * Set the visibility for the specified text
         * value.
         *
         * @param  text    Text value for which to set
         *                 the color.
         * @param  visible New visibility.
         * @return True if the value was set, false
         *         otherwise.
         */
        public boolean setVisibilityForText(String text, Boolean visible) {
            for (int j = 0; j < values.length; j++) {
                if (values[j].equals(text)) {
                    visibilityFlags[j] = visible;
                    return true;
                }
            }
            return false;
        }

        /**
         * Set the color for the specified text value.
         *
         * @param  text  Text value for which to set
         *               the color.
         * @param  color New color value.
         * @return True if the value was set, false
         *         otherwise.
         */
        public boolean setColorForText(String text, Color color) {
            for (int j = 0; j < values.length; j++) {
                if (values[j].equals(text)) {
                    colors[j] = color;
                    return true;
                }
            }
            return false;
        }
    }

    /**
     * Shapefile property of the integer (INTEGER) type.
     */
    public static class IntegerProperty extends Property {


        // Protected Variables

        /**
         * Minimum possible value.
         */
        protected int minimum;

        /**
         * Maximum possible value.
         */
        protected int maximum;

        /**
         * List of property value integers, starting at
         * the minimum value and going to the maximum;
         * each of these except the last has an associated
         * color at the corresponding index in <code>
         * colors</code>, meaning that the associated
         * color is to be used for all values between the
         * corresponding value (inclusive) and the next
         * index's value (exclusive, except for the last
         * value).
         */
        protected ArrayList<Integer> values = null;

        /**
         * List of visibility flags; these are paired with
         * the integer values at the corresponding indices
         * of <code>values</code>. Note that there is one
         * less element in this list than in <code>values
         * </code> because the last integer value does not
         * require a flag.
         */
        protected ArrayList<Boolean> visibilityFlags = null;

        /**
         * List of colors; these are paired with the
         * integer values at the corresponding indices of
         * <code>values</code>. Note that there is one
         * less element in this list than in <code>values
         * </code> because the last integer value does not
         * require a color.
         */
        protected ArrayList<Color> colors = null;


        // Public Constructors

        /**
         * Construct a standard instance with all integer
         * values visible and mapped to the same color.
         *
         * @param name   Property name.
         * @param values List of integer values; must be
         *               at least two elements in length,
         *               with each element being greater
         *               than the previous.
         * @param color  Starting color for all values.
         */
        public IntegerProperty(String name, int[] values, Color color) {
            super(name);
            this.values = new ArrayList<Integer>();
            this.visibilityFlags = new ArrayList<Boolean>();
            this.colors = new ArrayList<Color>();
            this.minimum = values[0];
            this.maximum = values[values.length - 1];
            for (int j = 0; j < values.length; j++) {
                this.values.add(values[j]);
                if (j < values.length - 1) {
                    this.visibilityFlags.add(Boolean.TRUE);
                    this.colors.add(color);
                }
            }
        }

        /**
         * Construct a standard instance.
         *
         * @param name     Property name.
         * @param values   List of integer values; must be
         *                 at least two elements in length,
         *                 with each element being greater
         *                 than the previous.
         * @param visFlags List of visibility flags for the
         *                 values; must be one less element
         *                 in length than <code>values</code>,
         *                 with each flag being valid for
         *                 all numbers between the integer
         *                 value at the corresponding index
         *                 in <code>values</code> (inclusive)
         *                 to the value at next index
         *                 (exclusive except for the last).
         * @param colors   List of colors for the integer
         *                 values; must be one less element
         *                 in length than <code>values</code>,
         *                 with each color being valid for
         *                 all numbers between the integer
         *                 value at the corresponding index
         *                 in <code>values</code> (inclusive)
         *                 to the value at next index
         *                 (exclusive except for the last).
         */
        public IntegerProperty(String name, int[] values,
                               Boolean[] visFlags, Color[] colors) {
            super(name);
            this.values = new ArrayList<Integer>();
            this.visibilityFlags = new ArrayList<Boolean>();
            this.colors = new ArrayList<Color>();
            this.minimum = values[0];
            this.maximum = values[values.length - 1];
            for (int j = 0; j < values.length; j++) {
                this.values.add(values[j]);
                if (j < colors.length) {
                    this.visibilityFlags.add(visFlags[j]);
                    this.colors.add(colors[j]);
                }
            }
        }


        // Public Methods

        /**
         * Return a clone of this object.
         *
         * @return Clone of this object.
         */
        @SuppressWarnings(value={"unchecked"})
			public Object clone() {
            IntegerProperty obj = null;
            try {
                obj = (IntegerProperty) super.clone();
            } catch (Exception e) {
                Logger.logBug("SHOULDN'T HAPPEN.", e);
            }
            obj.values = (ArrayList<Integer>) values.clone();
            obj.visibilityFlags = (ArrayList<Boolean>) visibilityFlags.clone();
            obj.colors = (ArrayList<Color>) colors.clone();
            return obj;
        }

        /**
         * Get the total number of integer ranges.
         *
         * @return Total number of integer ranges.
         */
        public int getSize() {
            return colors.size();
        }

        /**
         * Get the minimum integer value possible.
         *
         * @return Mimimum integer value possible.
         */
        public int getMinimum() {
            return minimum;
        }

        /**
         * Get the maximum integer value possible.
         *
         * @return Maximum integer value possible.
         */
        public int getMaximum() {
            return maximum;
        }

        /**
         * Get the integer range boundary at the specified
         * index.
         *
         * @param  index Index from which to fetch the
         *               integer range value; must be
         *               less than or equal to the value
         *               returned by <code>getSize()</code>.
         * @return Integer range boundary value.
         */
        public int getRangeBoundary(int index) {
            return values.get(index);
        }

        /**
         * Determine whether the specified index is
         * visible.
         *
         * @param  index Index for which to check the
         *               visibility.
         * @return Visibility of the index.
         */
        public Boolean isVisible(int index) {
            return visibilityFlags.get(index);
        }

        /**
         * Determine whether the specified value is
         * visible.
         *
         * @param  value Value for which to get
         *               the visibility.
         * @return Visibility for the value, or
         *         <code>null</code> if no such value
         *         is found.
         */
        public Boolean getVisibilityForValue(int value) {
            for (int j = 0; j < visibilityFlags.size(); j++) 
                if (value < values.get(j + 1))
                    return visibilityFlags.get(j);
            return visibilityFlags.get(visibilityFlags.size() - 1);
        }

        /**
         * Get the color at the specified index.
         *
         * @param  index Index from which to fetch the
         *               color; must be less than the value
         *               returned by <code>getSize()</code>.
         * @return Color.
         */
        public Color getColor(int index) {
            return colors.get(index);
        }

        /**
         * Get the color for the specified integer value.
         *
         * @param  value Integer value for which to get
         *               the color.
         * @return Color for the integer value.
         */
        public Color getColorForValue(int value) {
            for (int j = 0; j < colors.size(); j++) 
                if (value < values.get(j + 1))
                    return colors.get(j);
            return colors.get(colors.size() - 1);
        }

        /**
         * Determine whether the specified property
         * has the same name and type as this one.
         *
         * @param  other Other property to which this
         *               one is to be compared.
         * @return True if the properties have the
         *         same name and type, otherwise
         *         false.
         */
        public boolean hasSameNameAndTypeAs(Property other) {
            return (name.equals(other.name) && (other instanceof IntegerProperty));
        }

        /**
         * Determine whether the specified property
         * is equivalent to this one in terms of
         * its name, type, and potential values.
         *
         * @param  other Other property to which this
         *               one is to be compared.
         * @return True if the properties are
         *         equivalent as defined above, false
         *         otherwise.
         */
        public boolean isBasicallySameAs(Property other) {

            // If the name, type, or min and max values
            // are different, return false.
            if ((name.equals(other.name) == false) ||
                ((other instanceof IntegerProperty) == false))
                return false;
            IntegerProperty other2 = (IntegerProperty) other;
            if ((minimum != other2.minimum) || (maximum != other2.maximum))
                return false;
            return true;
        }

        /**
         * Get a list of the property values as
         * displayable strings.
         *
         * @return List of property values as strings
         *         suitable for display.
         */
        public ArrayList<String> getDisplayablePropertyValues() {
            ArrayList<String> list = new ArrayList<String>();
            for (int j = 0; j < values.size() - 1; j++) {
                int min = values.get(j);
                int max = values.get(j + 1);
                if (max - min == 1)
                    list.add(Integer.toString(min));
                else
                    list.add(min + " - " + (max - (j == values.size() - 2 ? 0 : 1)));
            }
            return list;
        }

        /**
         * Get a list of the visibility flags that
         * correspond to property values at the same
         * indices.
         *
         * @return List of visibility flags.
         */
        @SuppressWarnings(value={"unchecked"})
            public ArrayList<Boolean> getVisibilities() {
            return (ArrayList<Boolean>) visibilityFlags.clone();
        }

        /**
         * Get a list of the colors that correspond to
         * property values at the same indices.
         *
         * @return List of colors.
         */
        @SuppressWarnings(value={"unchecked"})
			public ArrayList<Color> getColors() {
            return (ArrayList<Color>) colors.clone();
        }

        /**
         * Incorporate relevant mappings from the
         * specified property into this one. Relevant
         * mappings are those attached to property
         * values that this property also may possess.
         *
         * @param other Other property from which
         *              relevant mappings must be
         *              taken.
         */
        public void incorporateRelevantMappings(Property other) {

            // Iterate through the other property's
            // mappings, attempting to use each one
            // in turn; any that do not apply because
            // the value does not exist in this
            // property will automatically be ignored.
            IntegerProperty other2 = (IntegerProperty) other;
            for (int j = 0; j < other2.values.size() - 1; j++)
                createNewRange(other2.values.get(j), other2.values.get(j + 1),
                               other2.visibilityFlags.get(j), other2.colors.get(j));
        }

        /**
         * Set the flag indicating whether the
         * specified index is visible.
         *
         * @param  index   Index for which to set the
         *                 visibility.
         * @param  visible New visibility.
         */
        public void setVisibility(int index, Boolean visible) {
            visibilityFlags.set(index, visible);
        }

        /**
         * Set the color at the specified index.
         *
         * @param index Index at which to set the color.
         * @param color New color value to be used.
         */
        public void setColor(int index, Color color) {
            colors.set(index, color);
        }

        /**
         * Set the visibility for the specified value.
         *
         * @param  text    Value for which to set the
         *                 color.
         * @param  visible New visibility.
         * @return True if the value was set, false
         *         otherwise.
         */
        public boolean setVisibilityForValue(int value, Boolean visible) {
            if ((value < minimum) || (value > maximum))
                return false;
            for (int j = 0; j < visibilityFlags.size(); j++) {
                if (values.get(j) == value) {
                    visibilityFlags.set(j, visible);
                    return true;
                }
            }
            return false;
        }

        /**
         * Set the color for the specified integer
         * range.
         *
         * @param  value Integer value that is the lower
         *               bound of the range for which to
         *               set the value.
         * @param  color New color value.
         * @return True if the value was set, false
         *         otherwise.
         */
        public boolean setColorForValue(int value, Color color) {
            if ((value < minimum) || (value > maximum))
                return false;
            for (int j = 0; j < colors.size(); j++) {
                if (values.get(j) == value) {
                    colors.set(j, color);
                    return true;
                }
            }
            return false;
        }

        /**
         * Add a new integer range. If this range
         * encompasses one or more complete preexisting
         * ranges, those ranges are replaced.
         *
         * @param  min     Integer value that is the lower
         *                 bound (inclusive) of the new
         *                 range.
         * @param  max     Integer value that is the upper
         *                 bound (exclusive, unless it is
         *                 the maximum value overall) of
         *                 the new range.
         * @param  visible Flag indicating whether the
         *                 new range should be visible.
         * @param  color   Color value for the new range.
         * @return True if the range was created, false
         *         otherwise.
         */
        public boolean createNewRange(int min, int max, Boolean visible,
                                      Color color) {

            // Make sure the mininum and maximum are
            // sanity checked.
            if ((min > maximum) || (max < minimum))
                return false;
            if (min < minimum)
                min = minimum;
            if (max > maximum)
                max = maximum;
            if (min >= max)
                return false;

            // Find the space where the minimum should
            // be inserted, and put it there.
            int j;
            Boolean lastFlag = null;
            Color lastColor = null;
            for (j = 0; j < values.size(); j++) {
                if (min < values.get(j)) {
                    lastFlag = visibilityFlags.get(j - 1);
                    lastColor = colors.get(j - 1);
                    values.add(j, min);
                    visibilityFlags.add(j, visible);
                    colors.add(j, color);
                    break;
                } else if (min == values.get(j)) {
                    lastFlag = visibilityFlags.set(j, visible);
                    lastColor = colors.set(j, color);
                    break;
                }
            }

            // Find the space where the maximum should
            // be inserted, and put it there.
            j++;
            while (j < values.size()) {
                if (max > values.get(j)) {
                    values.remove(j);
                    lastFlag = visibilityFlags.remove(j);
                    lastColor = colors.remove(j);
                } else if (max == values.get(j))
                    break;
                else {
                    values.add(j, max);
                    visibilityFlags.add(j, lastFlag);
                    colors.add(j, lastColor);
                    break;
                }
            }
            return true;
        }
    }

    /**
     * Shapefile property of the floating point (FLOATING)
     * type.
     */
    public static class FloatingProperty extends Property {


        // Protected Static Constants

        /**
         * Decimal formatter used to create displayable
         * strings representing the floating point
         * numbers.
         */
        protected static final DecimalFormat FORMATTER =
            new DecimalFormat("0.#######E0");


        // Protected Variables

        /**
         * Minimum possible value.
         */
        protected double minimum;

        /**
         * Maximum possible value.
         */
        protected double maximum;

        /**
         * List of property value floats, starting at
         * the minimum value and going to the maximum;
         * each of these except the last has an associated
         * color at the corresponding index in <code>
         * colors</code>, meaning that the associated
         * color is to be used for all values between the
         * corresponding value (inclusive) and the next
         * index's value (exclusive, except for the last
         * value).
         */
        protected ArrayList<Double> values = null;

        /**
         * List of visibility flags; these are paired with
         * the float values at the corresponding indices
         * of <code>values</code>. Note that there is one
         * less element in this list than in <code>values
         * </code> because the last float value does not
         * require a flag.
         */
        protected ArrayList<Boolean> visibilityFlags = null;

        /**
         * List of colors; these are paired with the
         * float values at the corresponding indices of
         * <code>values</code>. Note that there is one
         * less element in this list than in <code>values
         * </code> because the last float value does not
         * require a color.
         */
        protected ArrayList<Color> colors = null;


        // Public Constructors

        /**
         * Construct a standard instance with all float
         * values visible and mapped to the same color.
         *
         * @param name   Property name.
         * @param values List of float values; must be
         *               at least two elements in length,
         *               with each element being greater
         *               than the previous.
         * @param color  Starting color for all values.
         */
        public FloatingProperty(String name, double[] values, Color color) {
            super(name);
            this.values = new ArrayList<Double>();
            this.visibilityFlags = new ArrayList<Boolean>();
            this.colors = new ArrayList<Color>();
            this.minimum = values[0];
            this.maximum = values[values.length - 1];
            for (int j = 0; j < values.length; j++) {
                this.values.add(values[j]);
                if (j < values.length - 1) {
                    this.visibilityFlags.add(Boolean.TRUE);
                    this.colors.add(color);
                }
            }
        }

        /**
         * Construct a standard instance.
         *
         * @param name     Property name.
         * @param values   List of float values; must be
         *                 at least two elements in length,
         *                 with each element being greater
         *                 than the previous.
         * @param visFlags List of visibility flags for the
         *                 values; must be one less element
         *                 in length than <code>values</code>,
         *                 with each flag being valid for
         *                 all numbers between the float
         *                 value at the corresponding index
         *                 in <code>values</code> (inclusive)
         *                 to the value at next index
         *                 (exclusive except for the last).
         * @param color    List of colors for the float
         *                 values; must be one less element
         *                 in length than <code>values</code>,
         *                 with each color being valid for
         *                 all numbers between the float
         *                 value at the corresponding index
         *                 in <code>values</code> (inclusive)
         *                 to the value at next index
         *                 (exclusive except for the last).
         */
        public FloatingProperty(String name, double[] values,
                                Boolean[] visFlags, Color[] colors) {
            super(name);
            this.values = new ArrayList<Double>();
            this.visibilityFlags = new ArrayList<Boolean>();
            this.colors = new ArrayList<Color>();
            this.minimum = values[0];
            this.maximum = values[values.length - 1];
            for (int j = 0; j < values.length; j++) {
                this.values.add(values[j]);
                if (j < colors.length) {
                    this.visibilityFlags.add(visFlags[j]);
                    this.colors.add(colors[j]);
                }
            }
        }


        // Public Methods

        /**
         * Return a clone of this object.
         *
         * @return Clone of this object.
         */
        @SuppressWarnings(value={"unchecked"})
			public Object clone() {
            FloatingProperty obj = null;
            try {
                obj = (FloatingProperty) super.clone();
            } catch (Exception e) {
                Logger.logBug("SHOULDN'T HAPPEN.", e);
            }
            obj.values = (ArrayList<Double>) values.clone();
            obj.visibilityFlags = (ArrayList<Boolean>) visibilityFlags.clone();
            obj.colors = (ArrayList<Color>) colors.clone();
            return obj;
        }

        /**
         * Get the total number of float ranges.
         *
         * @return Total number of float ranges.
         */
        public int getSize() {
            return colors.size();
        }

        /**
         * Get the minimum float value possible.
         *
         * @return Mimimum float value possible.
         */
        public double getMinimum() {
            return minimum;
        }

        /**
         * Get the maximum float value possible.
         *
         * @return Maximum float value possible.
         */
        public double getMaximum() {
            return maximum;
        }

        /**
         * Get the float range boundary at the specified
         * index.
         *
         * @param  index Index from which to fetch the
         *               float range value; must be
         *               less than or equal to the value
         *               returned by <code>getSize()</code>.
         * @return Float range boundary value.
         */
        public double getRangeBoundary(int index) {
            return values.get(index);
        }

        /**
         * Determine whether the specified index is
         * visible.
         *
         * @param  index Index for which to check the
         *               visibility.
         * @return Visibility of the index.
         */
        public Boolean isVisible(int index) {
            return visibilityFlags.get(index);
        }

        /**
         * Determine whether the specified value is
         * visible.
         *
         * @param  value Value for which to get
         *               the visibility.
         * @return Visibility for the value, or
         *         <code>null</code> if no such value
         *         is found.
         */
        public Boolean getVisibilityForValue(double value) {
            for (int j = 0; j < visibilityFlags.size(); j++) 
                if (value < values.get(j + 1))
                    return visibilityFlags.get(j);
            return visibilityFlags.get(visibilityFlags.size() - 1);
        }

        /**
         * Get the color at the specified index.
         *
         * @param  index Index from which to fetch the
         *               color; must be less than the value
         *               returned by <code>getSize()</code>.
         * @return Color.
         */
        public Color getColor(int index) {
            return colors.get(index);
        }

        /**
         * Get the color for the specified float value.
         *
         * @param  value Float value for which to get
         *               the color.
         * @return Color for the float value.
         */
        public Color getColorForValue(double value) {
            for (int j = 0; j < colors.size(); j++) 
                if (value < values.get(j + 1))
                    return colors.get(j);
            return colors.get(colors.size() - 1);
        }

        /**
         * Determine whether the specified property
         * has the same name and type as this one.
         *
         * @param  other Other property to which this
         *               one is to be compared.
         * @return True if the properties have the
         *         same name and type, otherwise
         *         false.
         */
        public boolean hasSameNameAndTypeAs(Property other) {
            return (name.equals(other.name) && (other instanceof FloatingProperty));
        }

        /**
         * Determine whether the specified property
         * is equivalent to this one in terms of
         * its name, type, and potential values.
         *
         * @param  other Other property to which this
         *               one is to be compared.
         * @return True if the properties are
         *         equivalent as defined above, false
         *         otherwise.
         */
        public boolean isBasicallySameAs(Property other) {

            // If the name, type, or min and max values
            // are different, return false.
            if ((name.equals(other.name) == false) ||
                ((other instanceof FloatingProperty) == false))
                return false;
            FloatingProperty other2 = (FloatingProperty) other;
            if ((minimum != other2.minimum) || (maximum != other2.maximum))
                return false;
            return true;
        }

        /**
         * Get a list of the property values as
         * displayable strings.
         *
         * @return List of property values as strings
         *         suitable for display.
         */
        public ArrayList<String> getDisplayablePropertyValues() {
            ArrayList<String> list = new ArrayList<String>();
            for (int j = 0; j < values.size() - 1; j++) {
                if ((j == 0) && (values.size() - 1 > 1))
                    list.add("x < " + FORMATTER.format(values.get(j + 1)));
                else if ((j == values.size() - 2) && (values.size() - 1 > 1))
                    list.add("x \u2265 " + FORMATTER.format(values.get(j)));
                else
                    list.add(FORMATTER.format(values.get(j)) + " \u2264 x " +
                             (j == values.size() - 2 ? "\u2264 " : "< ") +
                             FORMATTER.format(values.get(j + 1)));
            }
            return list;
        }

        /**
         * Get a list of the visibility flags that
         * correspond to property values at the same
         * indices.
         *
         * @return List of visibility flags.
         */
        @SuppressWarnings(value={"unchecked"})
            public ArrayList<Boolean> getVisibilities() {
            return (ArrayList<Boolean>) visibilityFlags.clone();
        }

        /**
         * Get a list of the colors that correspond to
         * property values at the same indices.
         *
         * @return List of colors.
         */
        @SuppressWarnings(value={"unchecked"})
			public ArrayList<Color> getColors() {
            return (ArrayList<Color>) colors.clone();
        }

        /**
         * Incorporate relevant mappings from the
         * specified property into this one. Relevant
         * mappings are those attached to property
         * values that this property also may possess.
         *
         * @param other Other property from which
         *              relevant mappings must be
         *              taken.
         */
        public void incorporateRelevantMappings(Property other) {

            // Iterate through the other property's
            // mappings, attempting to use each one
            // in turn; any that do not apply because
            // the value does not exist in this
            // property will automatically be ignored.
            FloatingProperty other2 = (FloatingProperty) other;
            for (int j = 0; j < other2.values.size() - 1; j++)
                createNewRange(other2.values.get(j), other2.values.get(j + 1),
                               other2.visibilityFlags.get(j), other2.colors.get(j));
        }

        /**
         * Set the flag indicating whether the
         * specified index is visible.
         *
         * @param  index   Index for which to set the
         *                 visibility.
         * @param  visible New visibility.
         */
        public void setVisibility(int index, Boolean visible) {
            visibilityFlags.set(index, visible);
        }

        /**
         * Set the color at the specified index.
         *
         * @param index Index at which to set the color.
         * @param color New color value to be used.
         */
        public void setColor(int index, Color color) {
            colors.set(index, color);
        }

        /**
         * Set the visibility for the specified value.
         *
         * @param  text    Value for which to set the
         *                 color.
         * @param  visible New visibility.
         * @return True if the value was set, false
         *         otherwise.
         */
        public boolean setVisibilityForValue(double value, Boolean visible) {
            if ((value < minimum) || (value > maximum))
                return false;
            for (int j = 0; j < visibilityFlags.size(); j++) {
                if (values.get(j) == value) {
                    visibilityFlags.set(j, visible);
                    return true;
                }
            }
            return false;
        }

        /**
         * Set the color for the specified float
         * range.
         *
         * @param  value Float value that is the lower
         *               bound of the range for which to
         *               set the value.
         * @param  color New color value.
         * @return True if the value was set, false
         *         otherwise.
         */
        public boolean setColorForValue(double value, Color color) {
            if ((value < minimum) || (value > maximum))
                return false;
            for (int j = 0; j < colors.size(); j++) {
                if (values.get(j) == value) {
                    colors.set(j, color);
                    return true;
                }
            }
            return false;
        }

        /**
         * Add a new float range. If this range
         * encompasses one or more complete preexisting
         * ranges, those ranges are replaced.
         *
         * @param  min     Float value that is the lower
         *                 bound (inclusive) of the new
         *                 range.
         * @param  max     Float value that is the upper
         *                 bound (exclusive, unless it is
         *                 the maximum value overall) of
         *                 the new range.
         * @param  visible Flag indicating whether the
         *                 new range should be visible.
         * @param  color   Color value for the new range.
         * @return True if the range was created, false
         *         otherwise.
         */
        public boolean createNewRange(double min, double max, Boolean visible,
                                      Color color) {

            // Make sure the mininum and maximum are
            // sanity checked.
            if ((min > maximum) || (max < minimum))
                return false;
            if (min < minimum)
                min = minimum;
            if (max > maximum)
                max = maximum;
            if (min >= max)
                return false;

            // Find the space where the minimum should
            // be inserted, and put it there.
            int j;
            Boolean lastFlag = null;
            Color lastColor = null;
            for (j = 0; j < values.size(); j++) {
                if (min < values.get(j)) {
                    lastFlag = visibilityFlags.get(j - 1);
                    lastColor = colors.get(j - 1);
                    values.add(j, min);
                    visibilityFlags.add(j, visible);
                    colors.add(j, color);
                    break;
                } else if (min == values.get(j)) {
                    lastFlag = visibilityFlags.set(j, visible);
                    lastColor = colors.set(j, color);
                    break;
                }
            }

            // Find the space where the maximum should
            // be inserted, and put it there.
            j++;
            while (j < values.size()) {
                if (max > values.get(j)) {
                    values.remove(j);
                    lastFlag = visibilityFlags.remove(j);
                    lastColor = colors.remove(j);
                } else if (max == values.get(j))
                    break;
                else {
                    values.add(j, max);
                    visibilityFlags.add(j, lastFlag);
                    colors.add(j, lastColor);
                    break;
                }
            }
            return true;
        }
    }


    // Public Static Methods

    /**
     * Get the maximum number of values allowed for a
     * text property to be included as a potential
     * display property mapper.
     *
     * @return Maximum number of values allowed for a
     *         text property to be included, or -1
     *         if there is no limit.
     */
    public static int getMaximumNumberTextValues() {
        return MAX_TEXT_PROPERTY_VALUES;
    }


    // Protected Variables

    /**
     * Timestamp indicating when this object was
     * originally created.
     */
    protected long creationTime = -1L;

    /**
     * List of property to color mappings possible.
     */
    protected ArrayList<Property> propertyMappings = null;

    /**
     * Property to color mapping currently in use.
     */
    protected int currentPropertyMappingIndex = -1;

    /**
     * Flag indicating whether or not labels should
     * be displayed.
     */
    protected boolean showingLabels = true;

    /**
     * Label font name.
     */
    protected String fontName = "SansSerif";

    /**
     * Label font size.
     */
    protected int fontSize = 12;

    /**
     * Label font color.
     */
    protected Color fontColor = Color.white;

    /**
     * Flag indicating whether or not the label
     * font is bold.
     */
    protected boolean fontBold = true;

    /**
     * Flag indicating whether or not the label
     * font is italic.
     */
    protected boolean fontItalic = false;


    // Public Constructors

    /**
     * Construct a standard instance that is not
     * populated.
     */
    public ShapefileDisplayProperties() {
    }

    /**
     * Construct a standard instance with no color
     * mapping currently selected.
     *
     * @param creationTime     Timestamp indicating
     *                         when this object
     *                         was originally created.
     * @param propertyMappings Property mappings.
     */
    public ShapefileDisplayProperties(long creationTime,
                                      ArrayList<Property> propertyMappings) {
        this.creationTime = creationTime;
        this.propertyMappings = propertyMappings;
    }


    // Public Methods

    /**
     * Return a clone of this object.
     *
     * @return Clone of this object.
     */
    public Object clone() {
        ShapefileDisplayProperties obj = null;
        try {
            obj = (ShapefileDisplayProperties) super.clone();
        } catch (Exception e) {
            Logger.logBug("SHOULDN'T HAPPEN.", e);
        }
        obj.propertyMappings = new ArrayList<Property>();
        for (int j = 0; j < propertyMappings.size(); j++)
            obj.propertyMappings.add((Property) propertyMappings.get(j).clone());
        return obj;
    }

    /**
     * Find out whether or not this object has had
     * its basic property mappings filled in.
     *
     * @return True if the basic property mappings
     *         have been filled in, false otherwise.
     */
    public boolean isPopulated() {
        return (creationTime > -1);
    }

    /**
     * Get the timestamp indicating when this
     * object was originally created.
     *
     * @return Timestamp, or -1 if this object has
     *         not yet had its property mappings
     *         defined.
     */
    public long getCreationTime() {
        return creationTime;
    }

    /**
     * Get the total number of property mappings for
     * this file.
     *
     * @return Total number of property mappings, or
     *         -1 if this object has not yet had its
     *         property mappings defined.
     */
    public int getNumberOfPropertyMappings() {
        return (propertyMappings == null ? -1 : propertyMappings.size());
    }

    /**
     * Get the property mapping at the specified
     * index.
     *
     * @param  index Index of the property mapping
     *               to fetch.
     * @return Property mapping.
     */
    public Property getPropertyMapping(int index) {
        return (propertyMappings == null ? null : propertyMappings.get(index));
    }

    /**
     * Get all the property mappings.
     *
     * @return All property mappings.
     */
    public ArrayList<Property> getPropertyMappings() {
        return propertyMappings;
    }

    /**
     * Get the current property mappings index for
     * this file.
     *
     * @return Current property mappings index, or
     *         -1 if one has not yet been chosen.
     */
    public int getCurrentPropertyMappingsIndex() {
        return currentPropertyMappingIndex;
    }

    /**
     * Fill in the basic property mappings for this
     * object.
     *
     * @param creationTime     Timestmap indicating
     *                         when these mappings
     *                         were created.
     * @param propertyMappings Property mappings.
     * @param currentIndex     Currently selected
     *                         property mapping index.
     */
    public void populatePropertyMappings(long creationTime,
                                         ArrayList<Property> propertyMappings,
                                         int currentIndex) {
        this.creationTime = creationTime;
        this.propertyMappings = propertyMappings;
        this.currentPropertyMappingIndex = currentIndex;
    }

    /**
     * Copy the relevant property mappings and other
     * parameters from the given object to this one.
     * object.
     *
     * @param otherProperties Other properties object
     *                        from which parameters
     *                        are to be copied.
     */
    public void copyRelevantParameters(ShapefileDisplayProperties otherProperties) {
        creationTime = otherProperties.getCreationTime();
        propertyMappings = otherProperties.getPropertyMappings();
        currentPropertyMappingIndex = otherProperties.getCurrentPropertyMappingsIndex();
        showingLabels = otherProperties.isShowingLabels();
        fontName = otherProperties.getFontName();
        fontSize = otherProperties.getFontSize();
        fontColor = otherProperties.getFontColor();
        fontBold = otherProperties.isFontBold();
        fontItalic = otherProperties.isFontItalic();
    }

    /**
     * Determine whether the specified shapefile
     * has the same basic properties (that is, the
     * properties have the same names, types, and
     * potential values) as this shapefile.
     *
     * @param  other Other shapefile to which this
     *               one is to be compared.
     * @return True if the shapefiles are
     *         equivalent as defined above, false
     *         otherwise.
     */
    public boolean isBasicallySameAs(ShapefileDisplayProperties other) {

        // Ensure that the two shapefiles have the
        // same number of properties.
        if (propertyMappings.size() != other.propertyMappings.size())
            return false;

        // Iterate through the property mappings,
        // seeing if each has an equivalent in
        // the other shapefile's mappings.
        for (int j = 0; j < propertyMappings.size(); j++) {
            Property property = propertyMappings.get(j);
            boolean found = false;
            for (int k = 0; k < other.propertyMappings.size(); k++) {
                if (property.isBasicallySameAs(other.propertyMappings.get(k))) {
                    found = true;
                    break;
                }
            }
            if (found == false)
                return false;
        }
        return true;
    }

    /**
     * Incorporate any property mappings that are
     * are relevant (that is, mappings of those
     * properties that have semi-equivalent
     * properties) from the specified shapefile
     * into this one.
     *
     * @param other Shapefile from which mappings
     *              are to be incorporated.
     */
    public void incorporateRelevantMappings(ShapefileDisplayProperties other) {

        // Iterate through the property mappings,
        // seeing if each has an equivalent in
        // the other shapefile's mappings.
        for (int j = 0; j < propertyMappings.size(); j++) {
            Property property = propertyMappings.get(j);
            for (int k = 0; k < other.propertyMappings.size(); k++) {
                if (property.hasSameNameAndTypeAs(other.propertyMappings.get(k))) {
                    property.
                        incorporateRelevantMappings(other.propertyMappings.get(k));
                    if (k == other.currentPropertyMappingIndex)
                        currentPropertyMappingIndex = k;
                }
            }
        }

        // Reset the creation time to the other
        // properties' creation time.
        creationTime = other.creationTime;
    }

    /**
     * Make the specified property mapping the
     * current one.
     *
     * @param index Index of the property mapping
     *              to be made current.
     */
    public void setCurrentPropertyMapping(int index) {
        currentPropertyMappingIndex = index;
    }

    /**
     * Determine whether or not labels should be
     * displayed.
     *
     * @return True if labels should be displayed,
     *         false otherwise.
     */
    public boolean isShowingLabels() {
        return showingLabels;
    }

    /**
     * Set the flag indicating whether or not labels
     * should be displayed.
     *
     * @param showingLabels Flag indicating whether
     *                      or not labels should be
     *                      displayed.
     */
    public void setShowingLabels(boolean showingLabels) {
        this.showingLabels = showingLabels;
    }

    /**
     * Get the label font name.
     *
     * @return Label font name.
     */
    public String getFontName() {
        return fontName;
    }

    /**
     * Set the label font name.
     *
     * @param fontName Label font name.
     */
    public void setFontName(String fontName) {
        this.fontName = fontName;
    }

    /**
     * Get the label font size.
     *
     * @return Label font size.
     */
    public int getFontSize() {
        return fontSize;
    }

    /**
     * Set the label font size.
     *
     * @param fontSize Label font size.
     */
    public void setFontSize(int fontSize) {
        this.fontSize = fontSize;
    }

    /**
     * Get the label font color.
     *
     * @return Label font color.
     */
    public Color getFontColor() {
        return fontColor;
    }

    /**
     * Set the label font color.
     *
     * @param fontColor Label font color.
     */
    public void setFontColor(Color fontColor) {
        this.fontColor = fontColor;
    }

    /**
     * Determine whether or not the label font is
     * bold.
     *
     * @return True if the label font is bold,
     *         false otherwise.
     */
    public boolean isFontBold() {
        return fontBold;
    }

    /**
     * Set the flag indicating whether or not the
     * label font is bold.
     *
     * @param fontBold Flag indicating whether or
     *                 not the label font is bold.
     */
    public void setFontBold(boolean fontBold) {
        this.fontBold = fontBold;
    }

    /**
     * Determine whether or not the label font is
     * italic.
     *
     * @return True if the label font is italic,
     *         false otherwise.
     */
    public boolean isFontItalic() {
        return fontItalic;
    }

    /**
     * Set the flag indicating whether or not the
     * label font is italic.
     *
     * @param fontItalic Flag indicating whether
     *                   or not the label font is
     *                   italic.
     */
    public void setFontItalic(boolean fontItalic) {
        this.fontItalic = fontItalic;
    }
}
