package fsl.depict.shapefile;

import java.util.HashMap;
import fsl.depict.Depictable;
import fsl.depict.DepictableDisplayer;
import fsl.depict.DepictableDisplayProperties;
import fsl.depict.DepictableManager;
import fsl.gui.ComponentManager;
import fsl.gui.UserCommunicator;
import fsl.util.Logger;
import fsl.util.ProductInfo;
import fsl.util.StateHolder;

/**
 * The shapefile depictable manager, used to handle visual
 * properties and user manipulation of shapefile depictables.
 *
 * @author Christopher Golden
 * @see fsl.depict.shapefile.ShapefileDepictable
 */
public class ShapefileManager extends DepictableManager {


    // Public Static Constants

    /**
     * Change display property mappings command identifier.
     */
    public static final String CHANGE_DISPLAY_PROPERTIES = "shapefileChangeDisplayProperties";


    // Protected Static Constants

    /**
     * List of choice events handled by this class.
     */
    private static final String[] COMMANDS = { CHANGE_DISPLAY_PROPERTIES };

    /**
     * List of toggle events handled by this class.
     */
    private static final String[] TOGGLES = {};

    /**
     * List of choice events handled by this class.
     */
    private static final String[] CHOICES = {};


    // Protected Variables

    /**
     * Hash table mapping shapefile depict keys to
     * depictable display properties for any
     * shapefile that has been previously been
     * loaded.
     */
    protected HashMap<Long, ShapefileDisplayProperties> lastDisplayProperties;


    // Constructors

    /**
     * Construct a standard instance.
     *
     * @param depictDisplayer  Depictable displayer.
     * @param userCommunicator User communicator.
     * @param componentManager Component manager.
     * @param stateHolder      State holder.
     */
    @SuppressWarnings(value={"unchecked"})
    public ShapefileManager(DepictableDisplayer depictDisplayer,
                            UserCommunicator userCommunicator,
                            ComponentManager componentManager,
                            StateHolder stateHolder) {
        super(depictDisplayer, userCommunicator,
              componentManager, stateHolder);

        // Load any line thicknesses used the last
        // time any DGM map was loaded.
		lastDisplayProperties =
			(HashMap<Long, ShapefileDisplayProperties>) stateHolder.get(this, "shapefileDisplayProperties");
		if (lastDisplayProperties == null)
			lastDisplayProperties = new HashMap<Long, ShapefileDisplayProperties>();
    }


    // Public Methods

    /**
     * Create a display properties tracking object for the
     * specified map or product.
     *
     * @param  productInfo Product information
     * @param  isMap       Flag indicating whether or not the
     *                     product is a map.
     * @return Display properties tracking object.
     */
    public DepictableDisplayProperties
        createDisplayPropertiesTracker(ProductInfo productInfo,
                                       boolean isMap) {

        // Find the display properties that were saved before,
        // if any, or create a new empty set of properties
        // if none are found.
        ShapefileDisplayProperties properties = lastDisplayProperties.
            get(productInfo.getDepictKey());
        if (properties == null)
            properties = new ShapefileDisplayProperties();
        return properties;
    }

    /**
     * Prepare for the associated popup menu to be displayed.
     * 
     * @param depictable  Depictable for which the popup will
     *                    be displayed; this will be an
     *                    instance of the subclass of <code>
     *                    Depictable</code> associated with
     *                    this subclass by <code>
     *                    DepictableTypeInfo</code>.
     * @param productInfo Product information associated with
     *                    the depictable for which the popup
     *                    will be displayed.
     * @param properties  Display properties for the product
     *                    or map to which the depictable
     *                    belongs.
     * @param isMap       Flag indicating whether or not the
     *                    depictable for which the popup is
     *                    being displayed is considered a
     *                    map.
     * @param frameIndex  Frame index at which the depictable
     *                    is located.
     */
    public void prepareForPopupDisplay(Depictable depictable,
                                       ProductInfo productInfo,
                                       DepictableDisplayProperties properties,
                                       boolean isMap,
                                       int frameIndex) {
        /*
        componentManager.setChoice(THICKNESS_CHOICE,
                                   Integer.toString(((DGMDisplayProperties)
                                                     properties).getLineThickness()));
        */
    }

    /**
     * Determine whether or not the specified map or product
     * should allow itself to be converted to drawing glyphs.
     * 
     * @param depictable  Depictable for which the popup will
     *                    be displayed; this will be an
     *                    instance of the subclass of <code>
     *                    Depictable</code> associated with
     *                    this subclass by <code>
     *                    DepictableTypeInfo</code>.
     * @param productInfo Product information associated with
     *                    the depictable for which the popup
     *                    will be displayed.
     * @param properties  Display properties for the product
     *                    or map to which the depictable
     *                    belongs.
     * @param isMap       Flag indicating whether or not the
     *                    depictable for which the popup is
     *                    being displayed is considered a
     *                    map.
     * @param frameIndex  Frame index at which the depictable
     *                    is located.
     * @param allFrames   Flag indicating whether or not this
     *                    query pertains to converting all
     *                    depictables for this product or
     *                    map, or just the frame of the
     *                    specified depictable.
     */
    public boolean canBeConvertedToGlyphs(Depictable depictable,
                                          ProductInfo productInfo,
                                          DepictableDisplayProperties properties,
                                          boolean isMap, int frameIndex,
                                          boolean allFrames) {
        return false;
    }

    /**
     * Get the list of command events handled by this
     * class's <code>handleCommandEvent()</code> method.
     *
     * @return List of command events handled, or <code>
     *         null<code> if no such events are handled.
     */
    public String[] getCommandEventsHandled() {
        return COMMANDS;
    }

    /**
     * Get the list of toggle events handled by this
     * class's <code>handleToggleEvent()</code> method.
     *
     * @return List of toggle events handled, or <code>
     *         null<code> if no such events are handled.
     */
    public String[] getToggleEventsHandled() {
        return TOGGLES;
    }

    /**
     * Get the list of choice events handled by this
     * class's <code>handleChoiceEvent()</code> method.
     *
     * @return List of choice events handled, or <code>
     *         null<code> if no such events are handled.
     */
    public String[] getChoiceEventsHandled() {
        return CHOICES;
    }

	/**
	 * Handle the specified command event, triggered by a
     * menu item from the popup menu associated with this
     * class.
	 *
     * @param  depictable  Depictable for which the popup was
     *                     displayed; this will be an
     *                     instance of the subclass of <code>
     *                     Depictable</code> associated with
     *                     this subclass by <code>
     *                     DepictableTypeInfo</code>.
     * @param  productInfo Product information associated with
     *                     the depictable for which the popup
     *                     was displayed.
     * @param  properties  Display properties for the product
     *                     or map to which the depictable
     *                     belongs.
     * @param  isMap       Flag indicating whether or not the
     *                     depictable for which the popup was
     *                     displayed is considered a map.
     * @param  frameIndex  Frame index at which the depictable
     *                     is located.
	 * @param  identifier  Identifier of command to be handled.
	 * @param  parameter   Additional information that may be
	 *                     required, or <code>null</code>.
	 * @return True if the command has been handled, otherwise
	 *         false.
	 */
	public boolean handleCommandEvent(Depictable depictable, ProductInfo productInfo,
                                      DepictableDisplayProperties properties,
                                      boolean isMap, int frameIndex, String identifier,
                                      Object parameter) {

        // Handle the change display properties command, and
        // ignore everything else.
        if (identifier.equals(CHANGE_DISPLAY_PROPERTIES)) {
            ShapefileDisplayProperties oldProperties = (ShapefileDisplayProperties)
                properties;
            ShapefileDisplayProperties newProperties = (ShapefileDisplayProperties)
                oldProperties.clone();
            ShapefileDisplayPropertiesContent content =
                new ShapefileDisplayPropertiesContent(userCommunicator,
                                                      depictable.getLegend(),
                                                      newProperties,
                                                      depictDisplayer.getPalette());
            if (userCommunicator.getUserInputViaDialog("Shapefile Display Properties",
                                                       content)) {

                // If the size of the annotations may have
                // changed, tell the depictable this before
                // proceeding so that it can recalculate
                // its annotations and their positions.
                if ((oldProperties.getFontName().equals(newProperties.getFontName()) == false) ||
                    (oldProperties.getFontSize() != newProperties.getFontSize()) ||
                    (oldProperties.isFontBold() != newProperties.isFontBold()) ||
                    (oldProperties.isFontItalic() != newProperties.isFontItalic()))
                    depictable.annotationSizesChanged();

                // Remember the new display properties for
                // later.
                oldProperties.copyRelevantParameters(newProperties);
                lastDisplayProperties.put(productInfo.getDepictKey(), oldProperties);
                stateHolder.add(this, "shapefileDisplayProperties", lastDisplayProperties);
                stateHolder.save();

                // Tell the displayer to redraw all the
                // depictables for this map or product.
                depictDisplayer.redrawDepictable(productInfo.getDepictKey(),
                                                 isMap, -1, true);
            }
            return true;
        } else
            return false;
    }

	/**
	 * Handle the specified toggle event, triggered by a check
	 * box menu item from the popup menu associated with this
     * class.
	 *
     * @param  depictable  Depictable for which the popup was
     *                     displayed; this will be an
     *                     instance of the subclass of <code>
     *                     Depictable</code> associated with
     *                     this subclass by <code>
     *                     DepictableTypeInfo</code>.
     * @param  productInfo Product information associated with
     *                     the depictable for which the popup
     *                     was displayed.
     * @param  properties  Display properties for the product
     *                     or map to which the depictable
     *                     belongs.
     * @param  isMap       Flag indicating whether or not the
     *                     depictable for which the popup was
     *                     displayed is considered a map.
     * @param  frameIndex  Frame index at which the depictable
     *                     is located.
	 * @param  identifier  Identifier of switch that was
     *                     toggled.
	 * @param  state       New state of switch.
	 * @param  parameter   Additional information that may be
	 *                     required, or <code>null</code>.
	 * @return True if the toggle has been handled, otherwise
	 *         false.
	 */
	public boolean handleToggleEvent(Depictable depictable, ProductInfo productInfo,
                                     DepictableDisplayProperties properties,
                                     boolean isMap, int frameIndex, String identifier,
                                     boolean state, Object parameter) {
        return false;
    }

	/**
	 * Handle the specified choice event, triggered by a
	 * radio button menu item from the popup menu associated
     * with this class.
	 *
     * @param  depictable  Depictable for which the popup was
     *                     displayed; this will be an
     *                     instance of the subclass of <code>
     *                     Depictable</code> associated with
     *                     this subclass by <code>
     *                     DepictableTypeInfo</code>.
     * @param  productInfo Product information associated with
     *                     the depictable for which the popup
     *                     was displayed.
     * @param  properties  Display properties for the product
     *                     or map to which the depictable
     *                     belongs.
     * @param  isMap       Flag indicating whether or not the
     *                     depictable for which the popup was
     *                     displayed is considered a map.
     * @param  frameIndex  Frame index at which the depictable
     *                     is located.
	 * @param  identifier  Identifier providing the choice
	 *                     grouping.
	 * @param  chosen      String describing the choice made.
	 * @param  parameter   Additional information that may be
	 *                     required, or <code>null</code>.
	 * @return True if the choice has been handled, otherwise
	 *         false.
	 */
	public boolean handleChoiceEvent(Depictable depictable, ProductInfo productInfo,
                                     DepictableDisplayProperties properties,
                                     boolean isMap, int frameIndex, String identifier,
                                     String chosen, Object parameter) {
        return false;
    }
}